"""
This module lists all externally useful classes and functions
"""

from .data_request import WmsRequest, WcsRequest, AwsTileRequest, AwsProductRequest, \
    get_safe_format, download_safe_format

from .aws import AwsProduct, AwsTile
from .aws_safe import SafeProduct, SafeTile

from .common import BBox
from .constants import CRS, CustomUrlParam, MimeType, OgcConstants, AwsConstants

from .download import DownloadRequest, download_data, get_json, get_xml
from .opensearch import get_tile_info_id, get_tile_info, get_area_dates, get_area_info

from .io_utils import read_data, write_data
from .os_utils import get_content_list, get_folder_list, get_file_list, make_folder, create_parent_folder, rename, size
from .geo_utils import bbox_to_resolution, get_utm_bbox, wgs84_to_utm, to_wgs84, utm_to_wgs84, utm_to_pixel, \
    pixel_to_utm, wgs84_to_pixel, get_utm_epsg_from_latlon
from .time_utils import next_date, prev_date, get_current_date


__version__ = "1.0.0"



"""
Module for working with Sentinel Hub OGC services
"""

import logging
import datetime

from urllib.parse import urlencode

from .time_utils import get_current_date, parse_time
from .opensearch import get_area_dates
from .download import DownloadRequest
from .constants import CRS, DataSource, MimeType, OgcConstants
from .config import SGConfig

LOGGER = logging.getLogger(__name__)


class OgcService:
    """ Sentinel Hub OGC service class

    Intermediate layer between QGC-type requests (WmsRequest and WcsRequest) and the Sentinel Hub OGC (WMS and WCS)
    services.

    :param base_url: base url of Sentinel Hub's OGC services. If `None`, the url specified in the configuration
                    file is taken.
    :type base_url: None or str
    :param instance_id: user's instance id granting access to Sentinel Hub's OGC services. If `None`, the instance id
                        specified in the configuration file is taken.
    :type instance_id: None or str
    """
    def __init__(self, base_url=None, instance_id=None):
        self.base_url = SGConfig().ogc_base_url if base_url is None else base_url
        self.instance_id = SGConfig().instance_id if instance_id is None else instance_id
        if not self.instance_id:
            raise ValueError('Instance ID is not set. '
                             'Set it either in request initialisation or in configuration file.')

    def get_request(self, request):
        """ Get download requests

        Create a list of DownloadRequests for all Sentinel-2 acquisitions within request's time interval and
        acceptable cloud coverage.

        :param request: QGC request with specified bounding box, time interval, and cloud coverage for specific product.
        :type request: OgcRequest
        :return: list of DownloadRequests
        """
        return [DownloadRequest(url=self.get_url(request, date), filename=self.get_filename(request, date),
                                data_type=request.image_format) for date in self.get_dates(request)]

    def get_url(self, request, date):
        """ Returns url to Sentinel Hub's OGC service for the product specified by the OgcRequest and date.

        :param request: OGC request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest
        :param date: acquisition date of this product
        :type date: datetime
        :return: url to Sentinel Hub's OGC service for this product.
        """

        LOGGER.debug('MimeType: %s', request.image_format)

        url = self.base_url+str(request.source.value)

        params = {'SERVICE': str(request.source.value),
                  'BBOX': str(request.bbox),
                  'FORMAT': MimeType.get_string(request.image_format),
                  'CRS': CRS.ogc_string(request.bbox.get_crs()),
                  'MAXCC': 100.0 * request.maxcc}

        if request.source is DataSource.WMS:
            params = {**params,
                      **{
                          'WIDTH': request.size_x,
                          'HEIGHT': request.size_y,
                          'LAYERS': request.layer,
                          'REQUEST': 'GetMap'
                      }}
        elif request.source is DataSource.WCS:
            params = {**params,
                      **{
                          'RESX': request.size_x,
                          'RESY': request.size_y,
                          'COVERAGE': request.layer,
                          'REQUEST': 'GetCoverage'
                      }}

        if date is not None:
            start_date = date if request.time_difference < datetime.timedelta(
                seconds=0) else date - request.time_difference
            end_date = date if request.time_difference < datetime.timedelta(
                seconds=0) else date + request.time_difference
            params = {**params,
                      **{'TIME': '{0}/{1}'.format(start_date.isoformat(), end_date.isoformat())}}

        if request.custom_url_params is not None:
            params = {**params,
                      **{k.value: str(v) for k, v in request.custom_url_params.items()}}

        return '{0}/{1}?{2}'.format(url, self.instance_id, urlencode(params))

    @staticmethod
    def get_filename(request, date):
        """ Get filename location

        Returns the filename's location on disk where data is or is going to be stored.
        The files are stored in the folder specified by the user when initialising OGC-type
        of request. The name of the file has the following structure:

        {source}_{layer}_{crs}_{bbox}_{time}_{size_x}X{size_y}_{custom_url_param}_{custom_url_param_val}.{image_format}

        In case of TIFF_d32f a '_tiff_depth32f' is added at the end of the filename (before format suffix)
        to differentiate it from 16-bit float tiff.

        :param request: OGC request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest
        :param date: acquisition date of this product in `YYYY-MM-DDThh:mm:ss` format
        :type date: str
        :return: filename for this request and date
        """
        bbox_str = str(request.bbox).replace(',', '_')
        suffix = str(request.image_format.value)
        fmt = ''
        if request.image_format is MimeType.TIFF_d32f:
            suffix = str(MimeType.TIFF.value)
            fmt = str(request.image_format.value).replace(';', '_')

        filename = '_'.join([str(request.source.value),
                             request.layer,
                             str(request.bbox.get_crs()).replace(':', ''),
                             bbox_str,
                             date.strftime("%Y-%m-%dT%H-%M-%S"),
                             str(request.size_x)+'X'+str(request.size_y)])

        LOGGER.debug("filename=%s", filename)

        if request.custom_url_params is not None:
            for param, value in request.custom_url_params.items():
                filename = '_'.join([filename, param.value, str(value).replace('/', '_')])

        if fmt:
            filename = '_'.join([filename, fmt])

        filename = '.'.join([filename, suffix])

        return filename

    @staticmethod
    def _parse_date_interval(time):
        """ Parses dates into common form

        Parses specified time into common form - tuple of start and end dates, i.e.:

        (2017-01-15:T00:00:00, 2017-01-16:T23:59:59)

        The parameter can have the following values/format, which will be parsed as:

        * None -> [s2_start_date from config.json, current date]
        * YYYY-MM-DD -> [YYYY-MM-DD:T00:00:00, YYYY-MM-DD:T23:59:59]
        * YYYY-MM-DDThh:mm:ss -> [YYYY-MM-DDThh:mm:ss, YYYY-MM-DDThh:mm:ss]
        * list or tuple of two dates (YYYY-MM-DD) -> [YYYY-MM-DDT00:00:00, YYYY-MM-DDT23:59:59], where the first
          (second) element is start (end) date
        * list or tuple of two dates (YYYY-MM-DDThh:mm:ss) -> [YYYY-MM-DDThh:mm:ss, YYYY-MM-DDThh:mm:ss],
          where the first (second) element is start (end) date

        :param time: time window of acceptable acquisitions. See above for all acceptable argument formats.
        :type time: None, str of form `YYYY-MM-DD` or 'YYYY-MM-DDThh:mm:ss', list or tuple of two such strings
        :return: interval of start and end date of the form YYYY-MM-DDThh:mm:ss
        :rtype: tuple of start and end date
        """
        if time is None or time is OgcConstants.LATEST:
            date_interval = (SGConfig().s2_start_date, get_current_date())
        else:
            if isinstance(time, str):
                date_interval = (parse_time(time), parse_time(time))
            elif isinstance(time, list) or isinstance(time, tuple) and len(time) == 2:
                date_interval = (parse_time(time[0]), parse_time(time[1]))
            else:
                raise TabError('time must be a string or tuple of 2 strings or list of 2 strings')
            if date_interval[0] > date_interval[1]:
                raise ValueError('First time must be smaller or equal to second time')

        if len(date_interval[0].split('T')) == 1:
            date_interval = (date_interval[0] + 'T00:00:00', date_interval[1])
        if len(date_interval[1].split('T')) == 1:
            date_interval = (date_interval[0], date_interval[1] + 'T23:59:59')

        return date_interval

    @staticmethod
    def get_dates(request):
        """ Get available Sentinel-2 acquisitions at least time_difference apart

        List of all available Sentinel-2 acquisitions for given bbox with max cloud coverage and the specified
        time interval. When a single time is specified the request will return that specific date, if it exists.
        If a time range is specified the result is a list of all scenes between the specified dates conforming to
        the cloud coverage criteria. Most recent acquisition being first in the list.

        When a time_difference threshold is set to a positive value, the function filters out all datetimes which
        are within the time difference. The oldest datetime is preserved, all others all deleted.

        :param request: OGC-type request
        :type request: WmsRequest or WcsRequest
        """
        LOGGER.debug('CRS=%s', request.bbox.get_crs())

        date_interval = OgcService._parse_date_interval(request.time)

        LOGGER.debug('date_interval=%s', date_interval)

        if request.time is OgcConstants.LATEST:
            return OgcService._filter_dates(get_area_dates(request.bbox, date_interval, maxcc=request.maxcc)[-1:],
                                            request.time_difference)
        return OgcService._filter_dates(get_area_dates(request.bbox, date_interval, maxcc=request.maxcc),
                                        request.time_difference)

    @staticmethod
    def _filter_dates(dates, time_difference):
        """
        Filters out dates within time_difference, preserving only the oldest date.

        :param dates: a list of datetime objects
        :param time_difference: a datetime.timedelta representing the time difference threshold
        :return: an ordered list of datetimes d1<=d2<=...<=dn such that d[i+1]-di > time_difference
        :rtype: list[datetime.datetime]
        """

        LOGGER.debug("dates=%s", dates)

        if len(dates) <= 1:
            return dates

        sorted_dates = sorted(dates)

        separate_dates = [sorted_dates[0]]
        for curr_date in sorted_dates[1:]:
            if curr_date - separate_dates[-1] > time_difference:
                separate_dates.append(curr_date)
        return separate_dates



"""
Module that collects configuration data from config.json
"""

import os.path
import json

# pylint: disable=E0203

class SGConfig:
    """ This is a singleton implementation of the sentinelhub configuration class.

    The class reads during its first initialisation the configurable settings from
    ./config.json file:

        - ogc_base_url: base url for Sentinel Hub's services (should not be changed by the user)
        - aws_base_url: base url for Sentinel-2 data on AWS (should not be changed by the user)
        - aws_website_url: base url for AWS' public Sentinel-2 image browser
        - instance_id: users' instance id. User can set it to his/hers instance id in config.json instead of specifying
          it explicitly every time he/she creates new ogc request.

    Usage in the code:

        * SGConfig().ogc_base_url
        * SGConfig().instance_id

    """
    class _SGConfig:
        """
        Private class.
        """
        def __init__(self):
            self.config_params = ['ogc_base_url', 'instance_id', 'aws_base_url',
                                  'aws_website_url', 'opensearch_url',
                                  'max_opensearch_records_per_query', 's2_start_date',
                                  'max_download_attempts', 'download_sleep_time']
            self._load_configuration()

        def _check_configuration(self, config):
            """
            Checks if configuration file has contains all keys.

            :param config: configuration dictionary read from config.json
            :type config: dict
            """

            for key in self.config_params:
                if key not in config:
                    raise ValueError('Configuration file does not contain %s key.' % key)
            if config['max_opensearch_records_per_query'] > 500:
                raise ValueError("Value of config parameter 'max_opensearch_records_per_query' must be at most 500")

        def _load_configuration(self):
            """
            Method reads and loads the configuration file.
            """
            sentinelhub_dir = os.path.dirname(__file__)
            config_file = os.path.join(sentinelhub_dir, 'config.json')

            if not os.path.isfile(config_file):
                raise IOError('Configuration file does not exist: %s' % os.path.abspath(config_file))

            with open(config_file, 'r') as cfg_file:
                config = json.load(cfg_file)
                self._check_configuration(config)

                for prop in config:
                    if prop in self.config_params:
                        setattr(self, prop, config[prop])

                if not self.instance_id:
                    self.instance_id = None

    instance = None

    def __init__(self):
        if not SGConfig.instance:
            SGConfig.instance = SGConfig._SGConfig()

    def __getattr__(self, name):
        return getattr(self.instance, name)



"""
Module for downloading data
"""

# pylint: disable=too-many-instance-attributes

import logging
import os
import time
import concurrent.futures
import requests
import json
import cv2
import numpy as np
import tifffile as tiff
from io import BytesIO
from xml.etree import ElementTree

from .os_utils import create_parent_folder
from .constants import MimeType, RequestType
from .config import SGConfig

LOGGER = logging.getLogger(__name__)


class DownloadFailedException(Exception):
    pass


class ImageDecodingError(Exception):
    pass


class DownloadRequest:
    """ Class to manage HTTP requests

    Container for all download requests issued by the DataRequests containing
    url to Sentinel Hub's services or other sources to data, file names to
    saved data and other necessary flags needed when the data is downloaded and
    interpreted.

    :param url: url to Sentinel Hub's services or other sources from where the data is downloaded. Default is `None`
    :type url: str
    :param data_folder: folder name where the fetched data will be (or already is) saved. Default is `None`
    :type data_folder: str
    :param filename: filename of the file where the fetched data will be (or already is) saved. Default is `None`
    :type filename: str
    :param headers: add HTTP headers to request. Default is `None`
    :type headers: dict
    :param request_type: type of request, either GET or POST. Default is `constants.RequestType.GET`
    :type request_type: constants.RequestType
    :param post_values: form encoded data to send in POST request. Default is `None`
    :type post_values: dict
    :param save_response: flag to turn on/off saving data downloaded by this request to disk. Default is `True`.
    :type save_response: bool
    :param return_data: flag to return or not data downloaded by this request to disk. Default is `True`.
    :type return_data: bool
    :param data_type: expected file format of downloaded data. Default is `constants.MimeType.RAW`
    :type data_type: constants.MimeType

    """
    def __init__(self, *, url=None, data_folder=None, filename=None, headers=None, request_type=RequestType.GET,
                 post_values=None, save_response=True, return_data=True, data_type=MimeType.RAW, **properties):

        self.url = url
        self.data_folder = data_folder
        self.filename = filename
        self.headers = headers
        self.post_values = post_values
        self.save_response = save_response
        self.return_data = return_data

        self.properties = properties

        self.request_type = RequestType(request_type)
        self.data_type = MimeType(data_type)

        self.will_download = True
        self.file_location = None
        self._set_file_location()

    def set_save_response(self, save_response):
        """
        Set save_response attribute

        :param save_response: flag to turn on/off saving data downloaded by this request to disk. Default is `True`.
        :return: bool
        """
        self.save_response = save_response

    def set_return_data(self, return_data):
        """
        Set return_data attribute

        :param return_data: flag to return or not data downloaded by this request to disk. Default is `True`.
        :return: bool
        """
        self.return_data = return_data

    def set_data_folder(self, data_folder):
        """
        Set data_folder attribute

        :param data_folder: folder name where the fetched data will be (or already is) saved.
        :return: str
        """
        self.data_folder = data_folder
        self._set_file_location()

    def _set_file_location(self):
        if self.data_folder is not None and self.filename is not None:
            self.file_location = os.path.join(self.data_folder, self.filename.lstrip('/'))

    def is_downloaded(self):
        """ Checks if data for this request has already been downloaded and is saved to disk.

        :return: returns True if data for this request has already been downloaded and is saved to disk.
        :rtype: bool
        """
        if self.data_folder is None:
            return False
        self._set_file_location()
        return os.path.exists(self.file_location)

    def get_filename(self):
        """ Returns the full filename.

        :return: full filename (data folder + filename)
        :rtype: str
        """
        if self.data_folder:
            return os.path.join(self.data_folder, self.filename)
        return None


def download_data(request_list, redownload=False, max_threads=None):
    """ Download all requested data or read data from disk, if already downloaded and available and redownload is
    not required.

    :param request_list: list of DownloadRequests
    :type request_list: list of DownloadRequests
    :param redownload: if `True`, download again the data, although it was already downloaded and is available
                        on the disk. Default is `False`.
    :type redownload: bool
    :param max_threads: number of threads to use. Default is max_threads=None (5*N where N = nr. of cores on
                        the system)
    :type max_threads: int
    :return: list of Futures holding downloaded data, where each element in the list corresponds to an element
                in the download request list.
    :rtype: list[concurrent.futures.Future]
    """
    _check_if_must_download(request_list, redownload)

    LOGGER.debug("Using max_threads=%s for %s requests", max_threads, len(request_list))

    with concurrent.futures.ThreadPoolExecutor(max_workers=max_threads) as executor:
        return [executor.submit(execute_download_request, request) for request in request_list]


def _check_if_must_download(request_list, redownload):
    """
    Updates request.will_download attribute of each request in request_list.

    **Note:** the function mutates the elements of the list!

    :param request_list: a list of DownloadRequest instances
    :type: list[DownloadRequest]
    :param redownload: tells whether to download the data again or not
    :type: bool
    """
    for request in request_list:
        request.will_download = (request.save_response or request.return_data) \
                                and (not request.is_downloaded() or redownload)


def execute_download_request(request):
    """ Executes download request.

    :param request: DownloadRequest to be executed
    :type request: DownloadRequest
    :return: downloaded data or None
    :rtype: numpy array, other possible data type or None
    :raises: DownloadFailedException
    """
    if request.save_response and request.data_folder is None:
        raise ValueError('Data folder is not specified. '
                         'Please give a data folder name in the initialisation of your request.')

    if not request.will_download:
        return None

    try_num = SGConfig().max_download_attempts
    response = None
    while try_num > 0:
        try:
            response = _do_request(request)
            response.raise_for_status()
            LOGGER.debug('Downloaded from %s', request.url)
            break
        except (requests.HTTPError, requests.ConnectionError, requests.ConnectTimeout, requests.ReadTimeout,
                requests.Timeout) as exception:
            LOGGER.info('Download attempt failed with %s', exception)
        try_num -= 1
        if try_num > 0:
            LOGGER.debug('Unsuccessful download from %s ... will retry in %ds',
                         request.url, SGConfig().download_sleep_time)
            time.sleep(SGConfig().download_sleep_time)
        else:
            LOGGER.info('Failed to download from %s', request.url)
            raise DownloadFailedException('Failed to download from {}'.format(request.url))

    _save_if_needed(request, response)

    if request.return_data:
        return decode_data(response, request.data_type)
    return None


def _do_request(request):
    """
    Executes download request
    :param request: A request
    :type: DownloadRequest
    :return: Response of the request
    """
    if request.request_type is RequestType.GET:
        return requests.get(request.url, headers=request.headers)
    elif request.request_type is RequestType.POST:
        return requests.post(request.url, data=json.dumps(request.post_values), headers=request.headers)
    raise ValueError('Invalid request type {}'.format(request.request_type))


def _save_if_needed(request, response):
    """
    Save data to disk, if requested by the user
    :param request: Download request
    :type: DownloadRequest
    :param response: Response object from requests module
    """
    if request.save_response:
        create_parent_folder(request.file_location)
        with open(request.file_location, 'wb') as file:
            file.write(response.content)


def decode_data(response, data_type):
    """ Interprets downloaded data and returns it.

    :param response: downloaded data (i.e. json, png, tiff, xml, zip, ... file)
    :type response: any of possible data types returned by the download request
    :param data_type: expected downloaded data type
    :type data_type: constants.MimeType
    :return: downloaded data
    :rtype: numpy array in case of image data type, or other possible data type
    :raises: ValueError
    """
    LOGGER.debug('data_type=%s', data_type)

    if data_type is MimeType.JSON:
        return response.json()
    elif MimeType.is_image_format(data_type):
        return decode_image(response.content, data_type)
    elif data_type is MimeType.RAW or data_type is MimeType.TXT:
        return response.content
    elif data_type is MimeType.REQUESTS_RESPONSE:
        return response
    elif data_type is MimeType.ZIP:
        return BytesIO(response.content)
    elif data_type is MimeType.XML or data_type is MimeType.GML or \
            data_type is MimeType.SAFE:
        return ElementTree.fromstring(response.content)

    raise ValueError('Unknown response data type {}'.format(data_type))


def decode_image(data, image_type):
    """ Decodes the image provided in various formats, i.e. png, 16-bit float tiff, 32-bit float tiff, jp2
        and returns it as an numpy array

    :param data: image in its original format
    :type data: any of possible image types
    :param image_type: expected image format
    :type image_type: constants.MimeType
    :return: image as numpy array
    :rtype: numpy array
    :raises: ImageDecodingError
    """
    if image_type is MimeType.TIFF or image_type is MimeType.TIFF_d32f:
        return tiff.imread(BytesIO(data))
    elif image_type is MimeType.JP2:
        raise ImageDecodingError('Decoding JP2 images is currently not supported')

    img_array = np.asarray(bytearray(data))
    img = cv2.imdecode(img_array, cv2.IMREAD_UNCHANGED)

    if img is None:
        raise ImageDecodingError('Unable to decode image')
    return img


def get_json(url, post_values=None, headers=None):
    """ Download request as JSON data type

    :param url: url to Sentinel Hub's services or other sources from where the data is downloaded
    :type url: str
    :param post_values: form encoded data to send in POST request. Default is `None`
    :type post_values: dict
    :param headers: add HTTP headers to request. Default is `None`
    :type headers: dict
    :return: request response as JSON instance
    :rtype: JSON instance or None
    :raises: RunTimeError
    """

    json_headers = {} if headers is None else headers.copy()

    if post_values is None:
        request_type = RequestType.GET
    else:
        request_type = RequestType.POST
        json_headers = {**json_headers, **{'Content-Type': MimeType.get_string(MimeType.JSON)}}

    request = DownloadRequest(url=url, headers=json_headers, request_type=request_type, post_values=post_values,
                              save_response=False, return_data=True, data_type=MimeType.JSON)
    try:
        return execute_download_request(request)
    except DownloadFailedException:
        raise RuntimeError('Failed to obtain json from %s' % url)
    except Exception as err:
        LOGGER.debug('Failed with an non-DownloadFailedException on url=%s with an exception %s', url, err)
        raise


def get_xml(url):
    """ Download request as XML data type

    :param url: url to Sentinel Hub's services or other sources from where the data is downloaded
    :type url: str
    :return: request response as XML instance
    :rtype: XML instance or None
    :raises: RunTimeError
    """
    request = DownloadRequest(url=url, request_type=RequestType.GET, save_response=False, return_data=True,
                              data_type=MimeType.XML)
    try:
        return execute_download_request(request)
    except DownloadFailedException:
        raise RuntimeError('Failed to obtain xml from {}'.format(url))
    except Exception as err:
        LOGGER.debug('Failed with an non-DownloadFailedException on url=%s with an exception %s', url, err)
        raise



"""Module implementing classes common to all modules of the package (such as the bounding box class).

Representing bounding box can be confusing. If a function expects bbox as a list of four coordinates,
is it [lat1, lon1, lat2, lon2]? Or is it something else? And what CRS does it expect? Is the CRS a
separate parameter?

In this module the BBox class provides the canonical representation of a BBox that all the functions and
classes of the sg_utils package use, solving these issues.

Available classes:
 - BBox, represent a bounding box in a given CRS
"""

from .constants import CRS


class BBox:
    """ Class representing a bounding box in a given CRS.

    Throughout the sentinelhub package this class serves as the canonical representation of a bounding
    box. It can instantiate itself from multiple representations:

        1) ((min_x,min_y),(max_x,max_y))
        2) (min_x,min_y,max_x,max_y)
        3) [min_x,min_y,max_x,max_y]
        4) [[min_x, min_y],[max_x,max_y]],
        5) 'min_x,min_y,max_x,max_y'
        6) {'min_x':min_x, 'max_x':max_x, 'min_y':min_y, 'max_y':max_y}
        7) bbox, where bbox is an instance of BBox

    :param bbox: a bbox in a number of representations.
    :param crs: Coordinate Reference System that bbox is in. Expect one of the constants from the const.CRS enum.
    :type: constants.CRS
    """
    def __init__(self, bbox, crs):
        x_fst, y_fst, x_snd, y_snd = BBox._to_tuple(bbox)
        self.min_x = min(x_fst, x_snd)
        self.max_x = max(x_fst, x_snd)
        self.min_y = min(y_fst, y_snd)
        self.max_y = max(y_fst, y_snd)
        self.crs = CRS(crs)

    def __iter__(self):
        return iter(self.get_lower_left() + self.get_upper_right())

    def get_lower_left(self):
        """ Returns the lower left vertex of the bounding box

        :return: min_x, min_y
        """
        return self.min_x, self.min_y

    def get_upper_right(self):
        """ Returns the upper right vertex of the bounding box

        :return: max_x, max_y
        """
        return self.max_x, self.max_y

    def get_crs(self):
        return self.crs

    def __repr__(self):
        return self.get_lower_left(), self.get_upper_right(), self.crs

    def __str__(self):
        return "{},{},{},{}".format(self.min_x, self.min_y, self.max_x, self.max_y)

    @staticmethod
    def _to_tuple(bbox):
        """ Converts the input bbox representation (see the constructor docstring for a list of valid representations)
            into a flat tuple

        :param bbox: A bbox in one of 7 forms listed in the class description.
        :return: A flat tuple of size
        :raises: TypeError
        """

        if isinstance(bbox, list):
            return BBox._tuple_from_list(bbox)
        elif isinstance(bbox, str):
            return BBox._tuple_from_str(bbox)
        elif isinstance(bbox, tuple):
            if len(bbox) == 2 and all([isinstance(p, tuple) for p in bbox]):
                return bbox[0] + bbox[1]
            elif len(bbox) == 4 and all([isinstance(p, float) for p in bbox]):
                return bbox
        elif isinstance(bbox, dict):
            return BBox._tuple_from_dict(bbox)
        elif isinstance(bbox, BBox):
            return BBox._tuple_from_bbox(bbox)
        raise TypeError('Invalid bbox representation')

    @staticmethod
    def _tuple_from_str(bbox):
        """ Parses a string of the form 'min_x,min_y,max_x,max_y' into a flat tuple

        :param bbox: str of the form 'min_x,min_y,max_x,max_y'
        :return: tuple (min_x,min_y,max_x,max_y)
        """
        return tuple([float(s) for s in bbox.split(",")])

    @staticmethod
    def _tuple_from_list(bbox):
        """ Converts a list representation of a bbox into a flat tuple representation.

        :param bbox: a list of the form [min_x,min_y,max_x,max_y] or form [[min_x,min_y], [max_x,max_y]]
        :return: tuple (min_x,min_y,max_x,max_y)
        :raises: TypeError
        """
        if len(bbox) == 4 and all([isinstance(x, float) for x in bbox]):
            return tuple(bbox)
        elif len(bbox) == 2 and all([isinstance(lst, list) for lst in bbox]):
            return BBox._tuple_from_list(bbox[0] + bbox[1])
        raise TypeError('Expected a valid list representation of a bbox')

    @staticmethod
    def _tuple_from_dict(bbox):
        """ Converts a dictionary representation of a bbox into a flat tuple representation

        :param bbox: a dict with keys "min_x, "min_y", "max_x", and "max_y"
        :return: tuple (min_x,min_y,max_x,max_y)
        :raises: KeyError
        """
        return bbox['min_x'], bbox['min_y'], bbox['max_x'], bbox['max_y']

    @staticmethod
    def _tuple_from_bbox(bbox):
        """ Converts a BBox instance into a tuple

        :param bbox: An instance of the BBox type
        :return: tuple (min_x,min_y,max_x,max_y)
        """
        return bbox.get_lower_left() + bbox.get_upper_right()



"""
Module for managing files and folders
"""

import os
import errno


def get_content_list(folder='.'):
    """ Get list of contents in input folder

    :param folder: input folder to list contents. Default is `'.'`
    :type folder: str
    :return: list of folder contents
    :rtype: list(str)
    """
    return os.listdir(folder)


def get_folder_list(folder='.'):
    """ Get list of sub-folders contained in input folder

    :param folder: input folder to list sub-folders. Default is `'.'`
    :type folder: str
    :return: list of sub-folders
    :rtype: list(str)
    """
    dir_list = get_content_list(folder)
    return [f for f in dir_list if not os.path.isfile(os.path.join(folder, f))]


def get_file_list(folder='.'):
    """ Get list of files contained in input folder

    :param folder: input folder to list files only. Default is `'.'`
    :type folder: str
    :return: list of files
    :rtype: list(str)
    """
    dir_list = get_content_list(folder)
    return [f for f in dir_list if os.path.isfile(os.path.join(folder, f))]


def create_parent_folder(filename):
    """ Create parent folder for input filename recursively

    :param filename: input filename
    :type filename: str
    :raises: error if folder cannot be created
    """
    path = os.path.dirname(filename)
    if path != '':
        make_folder(path)


def make_folder(path):
    """ Create folder at input path recursively

    Create a folder specified by input path if one
    does not exist already

    :param path: input path to folder to be created
    :type path: str
    :raises: os.error if folder cannot be created
    """
    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise ValueError('Specified folder is not writable: %s'
                                 '\nPlease check permissions or set a new valid folder.' % path)


def rename(old_path, new_path, edit_folders=True):
    """ Rename files or folders

    :param old_path: name of file or folder to rename
    :param new_path: name of new file or folder
    :param edit_folders: flag to allow recursive renaming of folders. Default is `True`
    :type old_path: str
    :type new_path: str
    :type edit_folders: bool
    """
    if edit_folders:
        os.renames(old_path, new_path)
    else:
        os.rename(old_path, new_path)


def size(pathname):
    """ Returns size of a file or folder in Bytes

    :param pathname: path to file or folder to be sized
    :type pathname: str
    :return: size of file or folder in Bytes
    :rtype: int
    :raises: os.error if file is not accessible
    """
    if os.path.isfile(pathname):
        return os.path.getsize(pathname)
    return sum([size('{}/{}'.format(pathname, name)) for name in get_content_list(pathname)])



"""
Module for creating .SAFE structure with data from AWS
"""

from .download import get_xml
from .constants import AwsConstants, EsaSafeType, MimeType
from .aws import AwsProduct, AwsTile


class SafeProduct(AwsProduct):
    """ Class inherits from aws.AwsProduct """
    def get_requests(self):
        """
        Creates product structure and returns list of files for download

        :return: list of download requests
        :rtype: list(download.DownloadRequest)
        """
        safe = self.get_safe_struct()

        self.download_list = []
        self.structure_recursion(safe, self.data_folder)
        self.sort_download_list()
        return self.download_list

    def get_safe_struct(self):
        """
        Describes a structure inside tile folder of ESA product .SAFE structure

        :return: nested dictionaries representing .SAFE structure
        :rtype: dict
        """
        safe = {}
        main_folder = self.get_main_folder()
        safe[main_folder] = {}

        product_url = self.get_product_url()

        safe[main_folder][AwsConstants.AUX_DATA] = {}

        safe[main_folder][AwsConstants.DATASTRIP] = {}
        datastrip_list = self.get_datastrip_list()
        for datastrip_folder, datastrip_url in datastrip_list:
            safe[main_folder][AwsConstants.DATASTRIP][datastrip_folder] = {}
            safe[main_folder][AwsConstants.DATASTRIP][datastrip_folder][AwsConstants.QI_DATA] = {}
            safe[main_folder][AwsConstants.DATASTRIP][datastrip_folder][
                self.get_datastrip_metadata_name(datastrip_folder)] = datastrip_url + '/metadata.xml'

        safe[main_folder][AwsConstants.GRANULE] = {}

        for tile_info in self.product_info['tiles']:
            tile_name, date, aws_index = self.url_to_tile(self.get_tile_url(tile_info))
            if self.tile_list is None or AwsTile.parse_tile_name(tile_name) in self.tile_list:
                tile_struct = SafeTile(tile_name, date, aws_index, data_folder=None, bands=self.bands,
                                       metafiles=self.metafiles).get_safe_struct()
                for tile_name, safe_struct in tile_struct.items():
                    safe[main_folder][AwsConstants.GRANULE][tile_name] = safe_struct

        safe[main_folder][AwsConstants.HTML] = {}  # AWS doesn't have this data
        safe[main_folder][AwsConstants.INFO] = {}  # AWS doesn't have this data

        safe[main_folder]['INSPIRE.xml'] = self.get_url('inspire')
        safe[main_folder]['manifest.safe'] = product_url + '/manifest.safe'
        safe[main_folder][self.get_product_metadata_name()] = self.get_url('metadata')
        if self.safe_type == EsaSafeType.OLD_SAFE_TYPE:
            safe[main_folder][edit_name(self.product_id, 'BWI') + '.png'] = product_url + '/preview.png'
        return safe

    def get_main_folder(self):
        """
        :return: name of main folder
        :rtype: str
        """
        return self.product_id + '.SAFE'

    def get_datastrip_list(self):
        """
        :return: list of datastrips folder names and urls from productInfo.json file
        :rtype: list((str, str))
        """
        datastrips = self.product_info['datastrips']
        return [(self.get_datastrip_name(datastrip['id']), self.base_url + datastrip['path'])
                for datastrip in datastrips]

    def get_datastrip_name(self, datastrip):
        """
        :param datastrip: name of datastrip
        :type datastrip: str
        :return: name of datastrip folder
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_SAFE_TYPE:
            return datastrip
        return '_'.join(datastrip.split('_')[4:9])

    def get_datastrip_metadata_name(self, datastrip_folder):
        """
        :param datastrip_folder: name of datastrip folder
        :type: str
        :return: name of datastrip metadata file
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_SAFE_TYPE:
            name = datastrip_folder.rsplit('_', 1)[0]
        if self.safe_type == EsaSafeType.COMPACT_SAFE_TYPE:
            name = 'MTD_DS'
        return '{}.{}'.format(name, MimeType.XML.value)

    def get_product_metadata_name(self):
        """
        :return: name of product metadata file
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_SAFE_TYPE:
            name = edit_name(self.product_id, 'MTD', 'SAFL1C')
        if self.safe_type == EsaSafeType.COMPACT_SAFE_TYPE:
            name = 'MTD_MSIL1C'
        return '{}.{}'.format(name, MimeType.XML.value)


class SafeTile(AwsTile):
    """ Class inherits from aws.AwsTile """
    def __init__(self, *args, **kwargs):
        super(SafeTile, self).__init__(*args, **kwargs)

        self.safe_type = self.get_safe_type()
        self.tile_id = self.get_tile_id()

    def get_requests(self):
        """
        Creates tile structure and returns list of files for download

        :return: list of download requests for
        :rtype: list(download.DownloadRequest)
        """
        safe = self.get_safe_struct()

        self.download_list = []
        self.structure_recursion(safe, self.data_folder)
        self.sort_download_list()
        return self.download_list

    def get_safe_struct(self):
        """
        Describes a structure inside tile folder of ESA product .SAFE structure

        :return: nested dictionaries representing .SAFE structure
        :rtype: dict
        """
        safe = {}
        main_folder = self.get_main_folder()
        safe[main_folder] = {}

        safe[main_folder][AwsConstants.AUX_DATA] = {}
        safe[main_folder][AwsConstants.AUX_DATA][self.get_aux_data_name()] = self.tile_url + '/auxiliary/ECMWFT'

        safe[main_folder][AwsConstants.IMG_DATA] = {}
        for band in self.bands:
            safe[main_folder][AwsConstants.IMG_DATA][self.get_img_name(band)] = self.get_url(band)
        if self.safe_type == EsaSafeType.COMPACT_SAFE_TYPE:
            safe[main_folder][AwsConstants.IMG_DATA][self.get_img_name('TCI')] = self.get_url('TCI')

        safe[main_folder][AwsConstants.QI_DATA] = {}
        safe[main_folder][AwsConstants.QI_DATA][self.get_gml_name('CLOUDS')] = self.get_gml_url('CLOUDS')
        for qi_type in AwsConstants.QI_LIST:
            for band in self.bands:
                safe[main_folder][AwsConstants.QI_DATA][self.get_gml_name(qi_type, band)] = self.get_gml_url(qi_type,
                                                                                                             band)
        safe[main_folder][AwsConstants.QI_DATA][self.get_preview_name()] = self.tile_url + '/preview.jp2'

        safe[main_folder][self.get_tile_metadata_name()] = self.get_url(AwsConstants.METADATA)

        return safe

    def get_safe_type(self):
        """
        Determines the type of ESA product. In 2016 ESA changed structure and naming of data. Therefore the class must
        distinguish between old product type and compact (new) product type.

        :return: type of ESA product
        :rtype: constants.EsaSafeType
        """
        if self.get_product_id().split('_')[1] == 'MSIL1C':
            return EsaSafeType.COMPACT_SAFE_TYPE
        return EsaSafeType.OLD_SAFE_TYPE

    def get_tile_id(self):
        """
        Creates ESA tile ID

        :return: ESA tile ID
        :rtype: str
        """
        tree = get_xml(self.get_url(AwsConstants.METADATA))

        tile_id = tree[0].find('TILE_ID').text
        if self.safe_type == EsaSafeType.COMPACT_SAFE_TYPE:
            info = tile_id.split('_')
            tile_id = '_'.join([info[3], info[-2], info[-3], self.get_sensing_time()])
        return tile_id

    def get_sensing_time(self):
        """
        :return: Exact tile sensing time
        :rtype: str
        """
        return self.tile_info['timestamp'].split('.')[0].replace('-', '').replace(':', '')

    def get_datatake_time(self):
        """
        :return: Exact time of datatake
        :rtype: str
        """
        return self.tile_info['productName'].split('_')[2]

    def get_main_folder(self):
        """
        :return: name of tile folder
        :rtype: str
        """
        return self.tile_id

    def get_tile_metadata_name(self):
        """
        :return: name of tile metadata file
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_SAFE_TYPE:
            name = edit_name(self.tile_id, 'MTD', delete_end=True)
        if self.safe_type == EsaSafeType.COMPACT_SAFE_TYPE:
            name = 'MTD_TL'
        return name + '.xml'

    def get_aux_data_name(self):
        """
        :return: name of auxiliary data file
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_SAFE_TYPE:
            # this is not correct, but we cannot reconstruct last two timestamps in auxiliary data file name
            # e.g. S2A_OPER_AUX_ECMWFT_EPA__20160120T231011_V20160103T150000_20160104T030000
            return 'AUX_ECMWFT'
        return 'AUX_ECMWFT'

    def get_img_name(self, band):
        """
        :param band: band name
        :type band: str
        :return: name of band image file
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_SAFE_TYPE:
            name = self.tile_id.rsplit('_', 1)[0] + '_' + band
        if self.safe_type == EsaSafeType.COMPACT_SAFE_TYPE:
            name = '_'.join([self.tile_id.split('_')[1], self.get_datatake_time(), band])
        return name + '.jp2'

    def get_gml_name(self, qi_type, band='B00'):
        """
        :param qi_type: type of quality indicator
        :type qi_type: str
        :param band: band name
        :type band: str
        :return: name of gml file
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_SAFE_TYPE:
            name = edit_name(self.tile_id, 'MSK', delete_end=True)
            name = name.replace('L1C_TL', qi_type)
            name += '_' + band + '_MSIL1C'
        if self.safe_type == EsaSafeType.COMPACT_SAFE_TYPE:
            name = 'MSK_' + qi_type + '_' + band
        return name + '.gml'

    def get_gml_url(self, qi_type, band='B00'):
        """
        :param qi_type: type of quality indicator
        :type qi_type: str
        :param band: band name
        :type band: str
        :return: location of gml file on AWS
        :rtype: str
        """
        return self.tile_url + '/qi/MSK_' + qi_type + '_' + band + '.gml'

    def get_preview_name(self):
        """
        :return: name of preview file
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_SAFE_TYPE:
            name = edit_name(self.tile_id, 'PVI', delete_end=True)
        if self.safe_type == EsaSafeType.COMPACT_SAFE_TYPE:
            name = '_'.join([self.tile_id.split('_')[1], self.get_datatake_time(), 'PVI'])
        return name + '.jp2'


def edit_name(name, code, add_code=None, delete_end=False):
    """
    Helping function for creating file names in .SAFE format

    :param name: initial string
    :type name: str
    :param code:
    :type code: str
    :param add_code:
    :type add_code: str or None
    :param delete_end:
    :type delete_end: bool
    :return: edited string
    :rtype: str
    """
    info = name.split('_')
    info[2] = code
    if add_code is not None:
        info[3] = add_code
    if delete_end:
        info.pop()
    return '_'.join(info)



"""
Module with enum constants and utm utils
"""
import itertools as it
import mimetypes
import utm

from enum import Enum


mimetypes.add_type('application/json', '.json')


class DataSource(Enum):
    """ Enum constant class for data source services

    Supported types are WMS, WCS, AWS
    """
    WMS = 'wms'
    WCS = 'wcs'
    AWS = 'aws'


class _Direction(Enum):
    """ Enum constant class to encode NORTH/SOUTH direction """
    NORTH = 'N'
    SOUTH = 'S'


def _get_utm_code(zone, direction):
    """ Get UTM code given a zone and direction

    Direction is encoded as NORTH=6, SOUTH=7, while zone is the UTM zone number zero-padded.
    For instance, the code 32604 is returned for zone number 4, north direction.

    :param zone: UTM zone number
    :type zone: int
    :param direction: Direction enum type
    :type direction: Enum
    :return: UTM code
    :rtype: str
    """
    dir_dict = {_Direction.NORTH: '6', _Direction.SOUTH: '7'}
    return '{}{}{}'.format('32', dir_dict[direction], str(zone).zfill(2))


def _get_utm_name_value_pair(zone, direction=_Direction.NORTH):
    """ Get name and code for UTM coordinates

    :param zone: UTM zone number
    :type zone: int
    :param direction: Direction enum type
    :type direction: Enum, optional (default=NORTH)
    :return: Name and code of UTM coordinates
    :rtype: str, str
    """
    name = 'UTM_{}{}'.format(zone, direction.value)
    epsg = _get_utm_code(zone, direction)
    return name, epsg


class _BaseCRS(Enum):
    """ Coordinate Reference System enumerate class """
    def __str__(self):
        return self.ogc_string(self)

    @staticmethod
    def get_utm_from_wgs84(lat, lng):
        """ Convert from WGS84 to UTM coordinate system

        :param lat: Latitude
        :type lat: float
        :param lng: Longitude
        :type lng: float
        :return: UTM coordinates
        :rtype: tuple
        """
        _, _, zone, _ = utm.from_latlon(lat, lng)
        direction = 'N' if lat >= 0 else 'S'
        return CRS['UTM_{}{}'.format(str(zone), direction)]

    @staticmethod
    def ogc_string(crs):
        """ Returns a string of the form authority:id representing the CRS.

        :param crs: An enum constant representing a coordinate reference system.
        :type crs: Enum constant
        :return: A string representation of the CRS.
        :rtype: str
        """
        return 'EPSG:' + CRS(crs).value

    @classmethod
    def has_value(cls, value):
        """ Tests whether CRS contains a constant defined with string `value`.

        :param value: The string representation of the enum constant.
        :type value: str
        :return: `True` if there exists a constant with string value `value`, `False` otherwise
        :rtype: bool
        """
        return any(value == item.value for item in cls)


# Look-up class with possible combinations of UTM zone and direction
CRS = _BaseCRS("CRS", dict(
    [_get_utm_name_value_pair(zone, direction) for zone, direction in it.product(range(1, 65), _Direction)] +
    [('WGS84', '4326'), ('POP_WEB', '3857')]
))


class CustomUrlParam(Enum):
    """ Enum class to represent supported custom url parameters of OGC services

    Supported parameters are `AtmFilter`, `EvalScriptUrl`, `Preview`, `Quality`, `Upsampling`,
    `Downsampling`.

    See http://sentinel-hub.com/develop/documentation/api/custom-url-parameters for more information.
    """
    ATMFILTER = 'AtmFilter'
    EVALSCRIPTURL = 'EvalScriptUrl'
    PREVIEW = 'Preview'
    QUALITY = 'Quality'
    UPSAMPLING = 'Upsampling'
    DOWNSAMPLING = 'Downsampling'

    @classmethod
    def has_value(cls, value):
        """ Tests whether CustomUrlParam contains a constant defined with a string `value`

        :param value: The string representation of the enum constant
        :type value: str
        :return: `True` if there exists a constant with a string value `value`, `False` otherwise
        :rtype: bool
        """
        return any(value.lower() == item.value.lower() for item in cls)

    @staticmethod
    def get_string(param):
        """ Get custom url parameter name as string

        :param param: CustomUrlParam enum constant
        :type param: Enum constant
        :return: String describing the file format
        :rtype: str
        """
        return param.value


class MimeType(Enum):
    """ Enum class to represent supported image file formats

    Supported file formats are TIFF, TIFF 32-bit float, PNG, JPEG, JPEG2000, JSON, CSV, ZIP, HDF5, XML, GML, RAW

    """
    TIFF_d32f = 'tiff;depth32f'
    TIFF = 'tiff'
    PNG = 'png'
    JPG = 'jpg'
    JP2 = 'jp2'
    JSON = 'json'
    CSV = 'csv'
    ZIP = 'zip'
    HDF = 'hdf'
    XML = 'xml'
    GML = 'gml'
    TXT = 'txt'
    RAW = 'raw'
    SAFE = 'safe'
    REQUESTS_RESPONSE = 'response'  # http://docs.python-requests.org/en/master/api/#requests.Response

    @staticmethod
    def canonical_extension(fmt_ext):
        """ Canonical extension of file format extension

        Converts the format extension fmt_ext into the canonical extension for that format. For example,
        canonical_extension('tif') = 'tiff'. Here we agree that the canonical extension for format F is F.value

        :param fmt_ext: A string representing an extension (e.g. 'txt', 'png', etc.)
        :type fmt_ext: str
        :return: The canonical form of the extension (e.g. if fmt_ext='tif' then we return 'tiff')
        :rtype: str
        """
        if MimeType.has_value(fmt_ext):
            return fmt_ext
        try:
            return {
                'tif': MimeType.TIFF.value,
                'jpeg': MimeType.JPG.value,
                'hdf5': MimeType.HDF.value,
                'h5': MimeType.HDF.value
            }[fmt_ext]
        except KeyError:
            raise ValueError('Data format .{} is not supported'.format(fmt_ext))

    @staticmethod
    def is_image_format(value):
        """ Checks whether file format is an image format

        :param value: File format
        :type value: str
        :return: `True` if file is in image format, `False` otherwise
        :rtype: bool
        """
        return value in frozenset([MimeType.TIFF, MimeType.TIFF_d32f, MimeType.PNG, MimeType.JP2, MimeType.JPG])

    @classmethod
    def has_value(cls, value):
        """ Tests whether MimeType contains a constant defined with string `value`

        :param value: The string representation of the enum constant
        :type value: str
        :return: `True` if there exists a constant with string value `value`, `False` otherwise
        :rtype: bool
        """
        return any(value == item.value for item in cls)

    @staticmethod
    def get_string(fmt):
        """ Get file format as string

        :param fmt: MimeType enum constant
        :type fmt: Enum constant
        :return: String describing the file format
        :rtype: str
        """
        if fmt is MimeType.TIFF_d32f:
            return 'image/tiff;depth=32f'
        elif fmt is MimeType.JP2:
            return 'image/jpeg2000'
        elif fmt is MimeType.RAW or fmt is MimeType.REQUESTS_RESPONSE:
            return fmt.value
        return mimetypes.types_map['.' + fmt.value]


class RequestType(Enum):
    """ Enum constant class for GET/POST request type """
    GET = 'GET'
    POST = 'POST'


class OgcConstants:
    """ Initialisation of constants used by OGC request.

        Constants are LATEST
    """
    LATEST = 'latest'


class AwsConstants:
    """ Initialisation of constants used by AWS classes

        Constants are BANDS, TILE_INFO, PRODUCT_INFO, METADATA, PREVIEW, QI_LIST, AUX_DATA, DATASTRIP, GRANULE, HTML,
        INFO, QI_DATA, IMG_DATA, INSPIRE, MANIFEST, DATASTRIP_FILE, FILEFORMATS, PRODUCT_METAFILES, TILE_FILES
    """
    BANDS = ['B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08', 'B8A', 'B09', 'B10', 'B11', 'B12']
    TILE_INFO = 'tileInfo'
    PRODUCT_INFO = 'productInfo'
    METADATA = 'metadata'
    PREVIEW = 'preview'
    QI_LIST = ['DEFECT', 'DETFOO', 'NODATA', 'SATURA', 'TECQUA']
    AUX_DATA = 'AUX_DATA'
    DATASTRIP = 'DATASTRIP'
    GRANULE = 'GRANULE'
    HTML = 'HTML'
    INFO = 'rep_info'
    QI_DATA = 'QI_DATA'
    IMG_DATA = 'IMG_DATA'
    INSPIRE = 'inspire'
    MANIFEST = 'manifest'
    DATASTRIP_FILE = 'datastrip/*/metadata'

    FILE_FORMATS = {METADATA: MimeType.XML, PREVIEW: MimeType.JPG, TILE_INFO: MimeType.JSON,
                    PRODUCT_INFO: MimeType.JSON, 'TCI': MimeType.JP2, INSPIRE: MimeType.XML,
                    'auxiliary/ECMWFT': MimeType.RAW, MANIFEST: MimeType.SAFE, 'qi/MSK_CLOUDS_B00': MimeType.GML,
                    DATASTRIP_FILE: MimeType.XML}
    for band in BANDS:
        FILE_FORMATS[band] = MimeType.JP2
        FILE_FORMATS['preview/' + band] = MimeType.JP2
        for quality_indicator in QI_LIST:
            FILE_FORMATS['qi/MSK_{}_{}'.format(quality_indicator, band)] = MimeType.GML

    PRODUCT_METAFILES = [METADATA, PRODUCT_INFO, INSPIRE, MANIFEST, DATASTRIP_FILE]
    TILE_FILES = []
    for filename in FILE_FORMATS:
        if filename not in [INSPIRE, MANIFEST, DATASTRIP_FILE]:
            TILE_FILES.append(filename)


class EsaSafeType(Enum):
    """ Enum constants class for ESA .SAFE type.

     Types are OLD_SAFE_TYPE or COMPACT_SAFE_TYPE
     """
    OLD_SAFE_TYPE = 'old_type'
    COMPACT_SAFE_TYPE = 'compact_type'



import click

from .data_request import get_safe_format, download_safe_format
from .download import download_data, DownloadRequest


# pylint: disable=unused-argument

@click.command()
@click.option('--product', help='Product ID input')
@click.option('--tile', nargs=2, help='Tile name and date input')
@click.option('-f', '--folder', default='.', help='Set download location')
@click.option('-r', '--redownload', is_flag=True, default=False, help='Redownload existing files')
@click.option('-i', '--info', is_flag=True, default=False, help='Return safe format structure')
@click.option('-e', '--entire', is_flag=True, default=False, help='Get entire product of specified tile')
@click.option('-b', '--bands', default=None, help='Comma separated list (no spaces) of bands to retrieve')
def aws(product, tile, folder, redownload, info, entire, bands):
    """Download Sentinel-2 data from Sentinel-2 on AWS to ESA SAFE format. Download uses multiple threads.

    \b
    Examples:
      sentinelhub.aws --product S2A_MSIL1C_20170414T003551_N0204_R016_T54HVH_20170414T003551 \n
      sentinelhub.aws --product S2A_MSIL1C_20170414T003551_N0204_R016_T54HVH_20170414T003551 -i \n
      sentinelhub.aws --product S2A_MSIL1C_20170414T003551_N0204_R016_T54HVH_20170414T003551 -f /home/ESA_Products \n
      sentinelhub.aws --product S2A_MSIL1C_20170414T003551_N0204_R016_T54HVH_20170414T003551 --bands B08,B11 \n
      sentinelhub.aws --tile T54HVH 2017-04-14 \n
      sentinelhub.aws --tile T54HVH 2017-04-14 -e \n
    """
    band_list = None if bands is None else bands.split(',')
    if info:
        if product is None:
            click.echo(get_safe_format(tile=tile, entire_product=entire))
        else:
            click.echo(get_safe_format(product_id=product))
    else:
        if product is None:
            download_safe_format(tile=tile, folder=folder, redownload=redownload,
                                 entire_product=entire, bands=band_list)
        else:
            download_safe_format(product_id=product, folder=folder, redownload=redownload, bands=band_list)


@click.command()
@click.argument('url')
@click.argument('filename', type=click.Path())
@click.option('-r', '--redownload', is_flag=True, default=False, help='Redownload existing files')
def download(url, filename, redownload):
    """Download Sentinel-2 data from Sentinel-2 on AWS to ESA SAFE format. Download uses multiple threads.

    \b
    Example:
    sentinelhub.download http://sentinel-s2-l1c.s3.amazonaws.com/tiles/54/H/VH/2017/4/14/0/metadata.xml home/example.xml
    """
    download_list = [DownloadRequest(url=url, data_folder='', filename=filename, save_response=True, return_data=False)]
    download_data(download_list, redownload=redownload)



"""
Module for obtaining data from Amazon Web Service
"""

from abc import ABC, abstractmethod
import logging
import os.path

from .download import DownloadRequest, get_json
from .opensearch import get_tile_info, get_tile_info_id
from .time_utils import parse_time
from .config import SGConfig
from .constants import AwsConstants, EsaSafeType, MimeType


LOGGER = logging.getLogger(__name__)


class AwsService(ABC):
    """ Amazon Web Service (AWS) base class

    :param data_folder: location of the directory where the fetched data will be saved.
    :type data_folder: str
    :param bands: List of Sentinel-2 bands for request. If parameter is set to None all bands will be used.
    :type bands: list(str) or None
    :param metafiles: List of additional metafiles available on AWS
                      (e.g. ['metadata', 'tileInfo', 'preview/B01', 'TCI']).
                      If parameter is set to None the list will be set automatically.
    :type metafiles: list(str) or None
    """
    def __init__(self, data_folder='', bands=None, metafiles=None):
        self.data_folder = data_folder
        self.bands = self.parse_bands(bands)
        self.metafiles = self.parse_metafiles(metafiles)

        self.base_url = SGConfig().aws_base_url
        self.download_list = []

    @abstractmethod
    def get_requests(self):
        raise NotImplementedError

    @staticmethod
    def parse_bands(band_input):
        """
        Parses class input and verifies band names

        :param band_input: class input parameter 'bands'
        :type band_input: str or list(str)
        :return: verified list of bands
        :rtype: list(str)
        """
        if band_input is None:
            return AwsConstants.BANDS
        if isinstance(band_input, str):
            band_list = band_input.split(',')
        elif isinstance(band_input, list):
            band_list = band_input.copy()
        else:
            raise ValueError('bands parameter must be a list or a string')
        band_list = [band.strip().split('.')[0] for band in band_list]
        if not set(band_list) <= set(AwsConstants.BANDS):
            raise ValueError('bands must be a subset of {}'.format(AwsConstants.BANDS))
        return band_list

    def parse_metafiles(self, metafile_input):
        """
        Parses class input and verifies metadata file names

        :param metafile_input: class input parameter 'metafiles'
        :type metafile_input: str or list(str)
        :return: verified list of metadata files
        :rtype: list(str)
        """
        if metafile_input is None:
            if self.__class__.__name__ == 'SafeProduct' or self.__class__.__name__ == 'SafeTile':
                return sorted(AwsConstants.FILE_FORMATS.keys())
            return []
        if isinstance(metafile_input, str):
            metafile_list = metafile_input.split(',')
        elif isinstance(metafile_input, list):
            metafile_list = metafile_input.copy()
        else:
            raise ValueError('metafiles parameter must be a list or a string')
        metafile_list = [metafile.strip().split('.')[0] for metafile in metafile_list]
        if not set(metafile_list) <= set(AwsConstants.FILE_FORMATS.keys()):
            raise ValueError('metafiles must be a subset of {}'.format(
                list(AwsConstants.FILE_FORMATS.keys())))
        return metafile_list

    @staticmethod
    def url_to_tile(url):
        """
        Extracts tile name, date and AWS index from tile url on AWS

        :param url: class input parameter 'metafiles'
        :type url: str
        :return: Name of tile, date and AWS index which uniquely identifies tile on AWS
        :rtype: (str, str, int)
        """
        info = url.strip('/').split('/')
        name = ''.join(info[-7: -4])
        date = '-'.join(info[-4: -1])
        return name, date, int(info[-1])

    def sort_download_list(self):
        """
        Method that sorts the list of download requests. Band images have priority before metadata files. If bands
        images or metadata files are specified with a list they will be sorted in the same order as in the list.
        Otherwise they will be sorted alphabetically (band B8A will be between B08 and B09).
        """
        def aws_sort_function(download_request):
            data_name = download_request.properties['data_name']
            tile_url = download_request.url.rsplit('.', 1)[0].rstrip(data_name).rstrip('/')
            if data_name in AwsConstants.BANDS:
                return 0, tile_url, self.bands.index(data_name)
            return 1, tile_url, self.metafiles.index(data_name)
        self.download_list.sort(key=aws_sort_function)

    def structure_recursion(self, struct, folder):
        """
        From nested dictionaries representing .SAFE structure it recursively extracts all the files that need to be
        downloaded and stores them into class attribute 'download_list'

        :param struct:
        :type struct: dict
        :param folder: name of folder where this structure will be saved
        :type folder: str
        """
        if not struct:
            # This happens if the folder is empty. In current package version empty folder will not be created
            return
        for name, substruct in struct.items():
            subfolder = os.path.join(folder, name)
            if not isinstance(substruct, dict):
                if substruct.split('/')[3] == 'products':
                    data_name = substruct.split('/', 8)[-1]
                    if '/' in data_name:
                        items = data_name.split('/')
                        data_name = '/'.join([items[0], '*', items[2]])
                else:
                    data_name = substruct.split('/', 11)[-1]
                if '.' in data_name:
                    data_type = MimeType(substruct.split('.')[-1])
                    data_name = data_name.rsplit('.', 1)[0]
                else:
                    data_type = MimeType.RAW
                if data_name in self.bands + self.metafiles:
                    self.download_list.append(DownloadRequest(url=substruct, filename=subfolder, data_type=data_type,
                                                              data_name=data_name))
            else:
                self.structure_recursion(substruct, subfolder)


class AwsProduct(AwsService):
    """ Service class for Sentinel-2 product on AWS

    :param product_id: ESA ID of the product
    :type product_id: str
    :param tile_list: list of tile names
    :type tile_list: list(str) or None
    :param data_folder: location of the directory where the fetched data will be saved.
    :type data_folder: str
    :param bands: List of Sentinel-2 bands for request. If parameter is set to None all bands will be used.
    :type bands: list(str) or None
    :param metafiles: List of additional metafiles available on AWS
                      (e.g. ['metadata', 'tileInfo', 'preview/B01', 'TCI']).
                      If parameter is set to None the list will be set automatically.
    :type metafiles: list(str) or None
    """
    def __init__(self, product_id, tile_list=None, **kwargs):
        super(AwsProduct, self).__init__(**kwargs)

        self.product_id = product_id.split('.')[0]
        self.tile_list = self.parse_tile_list(tile_list)

        self.safe_type = self.get_safe_type()
        self.date = self.get_date()
        self.product_url = self.get_product_url()
        self.product_info = get_json(self.get_url(AwsConstants.PRODUCT_INFO))

    @staticmethod
    def parse_tile_list(tile_input):
        """
        Parses class input and verifies band names

        :param tile_input: class input parameter 'tile_list'
        :type tile_input: str or list(str)
        :return: parsed list of tiles
        :rtype: list(str) or None
        """
        if tile_input is None:
            return None
        if isinstance(tile_input, str):
            tile_list = tile_input.split(',')
        elif isinstance(tile_input, list):
            tile_list = tile_input.copy()
        else:
            raise ValueError('tile_list parameter must be a list of tile names')
        tile_list = [AwsTile.parse_tile_name(tile_name) for tile_name in tile_list]
        return tile_list

    def get_requests(self):
        """
        Creates product structure and returns list of files for download

        :return: List of download requests
        :rtype: list(download.DownloadRequest)
        """
        self.download_list = [DownloadRequest(url=self.get_url(metafile), filename=self.get_filepath(metafile),
                                              data_type=AwsConstants.FILE_FORMATS[metafile], data_name=metafile) for
                              metafile in self.metafiles if metafile in AwsConstants.PRODUCT_METAFILES]

        tile_data_folder = os.path.join(self.data_folder, self.product_id)
        for tile_info in self.product_info['tiles']:
            tile_name, date, aws_index = self.url_to_tile(self.get_tile_url(tile_info))
            if self.tile_list is None or AwsTile.parse_tile_name(tile_name) in self.tile_list:
                self.download_list.extend(AwsTile(tile_name, date, aws_index, data_folder=tile_data_folder,
                                                  bands=self.bands, metafiles=self.metafiles).get_requests())
        self.sort_download_list()
        return self.download_list

    def get_safe_type(self):
        """ Determines the type of ESA product.

        In 2016 ESA changed structure and naming of data. Therefore the class must
        distinguish between old product type and compact (new) product type.

        :return: type of ESA product
        :rtype: constants.EsaSafeType
        """
        if self.product_id.split('_')[1] == 'MSIL1C':
            return EsaSafeType.COMPACT_SAFE_TYPE
        return EsaSafeType.OLD_SAFE_TYPE

    def get_date(self):
        """ Collects sensing date of the product.

        :return: Sensing date
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_SAFE_TYPE:
            name = self.product_id.split('_')[-2]
            date = [name[1:5], name[5:7], name[7:9]]
        elif self.safe_type == EsaSafeType.COMPACT_SAFE_TYPE:
            name = self.product_id.split('_')[2]
            date = [name[:4], name[4:6], name[6:8]]
        return '-'.join(date_part.lstrip('0') for date_part in date)

    def get_url(self, filename):
        """
        Creates url of file location on AWS

        :param filename: name of file
        :type filename: str
        :return: url of file location
        :rtype: str
        """
        if self.product_url is None:
            self.product_url = self.get_product_url()
        return '{}/{}.{}'.format(self.product_url, filename, AwsConstants.FILE_FORMATS[filename].value)

    def get_product_url(self):
        """
        Creates base url of product location on AWS

        :return: url of product location
        :rtype: str
        """
        return self.base_url + 'products/' + self.date.replace('-', '/') + '/' + self.product_id

    def get_tile_url(self, tile_info):
        """
        Collects tile url from productInfo.json file.

        :param tile_info: information about tile from productInfo.json
        :type tile_info: dict
        :return: url of tile location
        :rtype: str
        """
        return self.base_url + '/' + tile_info['path']

    def get_filepath(self, filename):
        """
        Creates file path for the file

        :param filename: name of
        :type filename: str
        :return: filename with path on disk
        :rtype: str
        """
        return '{}/{}/{}.{}'.format(self.data_folder, self.product_id, filename,
                                    AwsConstants.FILE_FORMATS[filename].value).replace(':', '.')


class AwsTile(AwsService):
    """ Service class for Sentinel-2 product on AWS

    :param tile: Tile name (e.g. 'T10UEV')
    :type tile: str
    :param time: Tile sensing time in ISO8601 format
    :type time: str
    :param aws_index: There exist Sentinel-2 tiles with the same tile and time parameter. Therefore each tile on AWS
                      also has an index which is visible in their url path. If aws_index is set to None the class will
                      try to find the index automatically. If there will be multiple choices it will choose the lowest
                      index and inform the user.
    :type aws_index: int or None
    :param data_folder: location of the directory where the fetched data will be saved.
    :type data_folder: str
    :param bands: List of Sentinel-2 bands for request. If parameter is set to None all bands will be used.
    :type bands: list(str) or None
    :param metafiles: List of additional metafiles available on AWS
                      (e.g. ['metadata', 'tileInfo', 'preview/B01', 'TCI']).
                      If parameter is set to None the list will be set automatically.
    :type metafiles: list(str) or None
    """
    def __init__(self, tile_name, time, aws_index=None, **kwargs):
        super(AwsTile, self).__init__(**kwargs)

        self.tile_name = self.parse_tile_name(tile_name)
        self.datetime = self.parse_datetime(time)
        self.date = self.datetime.split('T')[0]
        self.aws_index = aws_index

        LOGGER.debug('tile_name=%s, date=%s, bands=%s, metafiles=%s', self.tile_name, self.date,
                     self.bands, self.metafiles)

        self.aws_index = self.get_aws_index()
        self.tile_url = self.get_tile_url()
        self.tile_info = self.get_tile_info()
        if not self.tile_is_valid():
            raise ValueError('Cannot find data on AWS for specified tile, time and aws_index')

    @staticmethod
    def parse_tile_name(name):
        """
        Parses and verifies tile name.

        :param name: class input parameter 'tile_name'
        :type name: str
        :return: parsed tile name
        :rtype: str
        """
        tile_name = name.lstrip('T0')
        if len(tile_name) == 4:
            tile_name = '0' + tile_name
        if len(tile_name) != 5:
            raise ValueError('Invalid tile name {}'.format(name))
        return tile_name

    @staticmethod
    def parse_datetime(time):
        """
        Parses and verifies tile sensing time.

        :param time: tile sensing time
        :type time: str
        :return: tile sensing time in ISO8601 format
        :rtype: str
        """
        try:
            return parse_time(time)
        except Exception:
            raise ValueError('Time must be in format YYYY-MM-DD or YYYY-MM-DDTHH:MM:SS')

    def set_data_folder(self, data_folder):
        """
        Sets folder where data will be stored

        :param data_folder: name of folder
        :type data_folder: str
        """
        self.data_folder = data_folder

    def get_requests(self):
        """
        Creates tile structure and returns list of files for download

        :return: List of download requests for
        :rtype: list(download.DownloadRequest)
        """
        self.download_list = []
        for data_name in self.bands + self.metafiles:
            if data_name in AwsConstants.TILE_FILES:
                url = self.get_url(data_name)
                filename = self.get_filepath(data_name)
                self.download_list.append(DownloadRequest(url=url, filename=filename,
                                                          data_type=AwsConstants.FILE_FORMATS[data_name],
                                                          data_name=data_name))
        self.sort_download_list()
        return self.download_list

    def get_aws_index(self):
        """
        Returns tile index on AWS. If it was not set it collects it with Opensearch

        :return: Index of tile on AWS
        :rtype: int
        """
        if self.aws_index is not None:
            return self.aws_index
        tile_info = get_tile_info(self.tile_name, self.datetime)
        if tile_info is not None:
            return int(tile_info['properties']['s3Path'].split('/')[-1])
        raise ValueError('Cannot find aws_index for specified tile and time')

    def tile_is_valid(self):
        return self.tile_info is not None \
               and (self.datetime == self.date or self.datetime == self.parse_datetime(self.tile_info['timestamp']))

    def get_tile_info(self):
        """
        Collects basic info about tile from tileInfo.json

        :return: dictionary with tile information
        :rtype: dict
        """
        url = self.get_url('tileInfo')
        try:
            return get_json(url)
        except Exception as err:
            LOGGER.error('Download from url %s failed with %s', url, err)
            raise

    def get_url(self, filename):
        """
        Creates url of file location on AWS

        :param filename: name of file
        :type filename: str
        :return: url of file location
        :rtype: str
        """
        if self.tile_url is None or filename == AwsConstants.TILE_INFO:
            self.tile_url = self.get_tile_url()
        return '{}/{}.{}'.format(self.tile_url, filename, AwsConstants.FILE_FORMATS[filename].value)

    def get_tile_url(self):
        """
        Creates base url of tile location on AWS

        :return: url of tile location
        :rtype: str
        """
        url = self.base_url + 'tiles/' + self.tile_name[0:2].lstrip('0') + '/' + self.tile_name[2] + '/' \
            + self.tile_name[3:5] + '/'
        date_params = self.date.split('-')
        for param in date_params:
            url += param.lstrip('0') + '/'
        return url + str(self.aws_index)

    def get_filepath(self, filename):
        """
        Creates file path for the file

        :param filename: name of
        :type filename: str
        :return: filename with path on disk
        :rtype: str
        """
        return '{}/{},{},{}/{}.{}'.format(self.data_folder, self.tile_name, self.date, self.aws_index, filename,
                                          AwsConstants.FILE_FORMATS[filename].value).replace(':', '.')

    def get_product_id(self):
        """
        Obtains ESA ID of product which contains the tile

        :return: ESA ID of the product
        :rtype: str
        """
        return self.tile_info['productName']

    @staticmethod
    def tile_id_to_tile(tile_id):
        """
        :param tile_id: original ESA tile ID
        :type: str
        :return: tile name, sensing date and AWS index
        :rtype: (str, str, int)
        """
        tile_info = get_tile_info_id(tile_id)
        return AwsService.url_to_tile(tile_info['properties']['s3Path'])



"""
Module for manipulation of geographical information.
"""

# pylint: disable=unused-argument

import logging
import math
import pyproj

from .constants import CRS
from .common import BBox


LOGGER = logging.getLogger(__name__)


WGS84_PROJ = pyproj.Proj(init=CRS.ogc_string(CRS.WGS84))
EPSG3857_PROJ = pyproj.Proj(init=CRS.ogc_string(CRS.POP_WEB))

ERR = 0.1

# Semi-axes of WGS-84 geoidal reference
EARTH_WGS84_A = 6378137.0  # Major semiaxis [m]
EARTH_WGS84_B = 6356752.3  # Minor semiaxis [m]


def earth_radius_wgs84(lat):
    """ Returns earth's radius at a given latitude, according to the WGS-84 ellipsoid [m]

    :param lat: latitude in radians
    :type lat: float
    :return: Earth's radius at a given latitude
    :rtype: float

    """

    ad_lat = EARTH_WGS84_A * math.cos(lat)
    bd_lat = EARTH_WGS84_B * math.sin(lat)
    an_lat = EARTH_WGS84_A * ad_lat
    bn_lat = EARTH_WGS84_B * bd_lat

    return math.sqrt((an_lat*an_lat + bn_lat*bn_lat)/(ad_lat*ad_lat + bd_lat*bd_lat))


def bbox_to_resolution_wgs84(bbox, width, height):
    """ Calculates pixel resolution in meters for a given bbox, width and height.

    :param bbox: latitude and longitude of south/west and north/east points
    :type bbox: list of two tuples
    :param width: width of bounding box in pixels
    :type width: int
    :param height: height of bounding box in pixels
    :type height: int
    :return: resolution east-west at north and south, and resolution north-south in meters
    :rtype: float, float, float

    """

    x_mn, y_mn = bbox.get_lower_left()
    x_mx, y_mx = bbox.get_upper_right()

    res_ew_n = (math.radians(y_mx) - math.radians(y_mn)) * earth_radius_wgs84(
        math.radians(x_mn)) * math.cos(math.radians(x_mx))
    res_ew_s = (math.radians(y_mx) - math.radians(y_mn)) * earth_radius_wgs84(
        math.radians(x_mn)) * math.cos(math.radians(x_mn))
    res_ns = (math.radians(x_mx) - math.radians(x_mn)) * earth_radius_wgs84(math.radians(x_mn))

    return res_ew_n / width, res_ew_s / width, res_ns / height


def bbox_to_resolution(bbox, width, height):
    """ Calculates pixel resolution in meters for a given bbox of a given width and height in a given CRS.

    :param bbox: latitude and longitude of south/west and north/east points
    :type bbox: list of two tuples
    :param width: width of bounding box in pixels
    :type width: int
    :param height: height of bounding box in pixels
    :type height: int
    :return: resolution east-west at north and south, and resolution north-south in meters for given CRS
    :rtype: float, float, float
    :raises: ValueError if CRS is not supported
    """
    crs = bbox.get_crs()
    x_mn, y_mn = bbox.get_lower_left()
    x_mx, y_mx = bbox.get_upper_right()

    if crs is CRS.WGS84:
        return bbox_to_resolution_wgs84(bbox, width, height)
    elif crs is CRS.POP_WEB:
        bbox_wgs84 = BBox(
            [pyproj.transform(EPSG3857_PROJ, WGS84_PROJ, point[0], point[1]) for point in ((x_mn, y_mn), (x_mx, y_mx))],
            CRS.WGS84)
        return bbox_to_resolution_wgs84(bbox_wgs84, width, height)
    raise ValueError('Given CRS {} is not supported.'.format(crs))


def get_utm_bbox(img_bbox, transform):  # img_bbox = [row1, column1, row2, column2]
    """ Get UTM coordinates given a bounding box in pixels and a transform

    :param img_bbox: boundaries of bounding box in pixels as [row1, col1, row2, col2]
    :type img_bbox: list
    :param transform: georeferencing transform
    :type transform: list
    :return: UTM coordinates as [east1, north1, east2, north2]
    :rtype: list
    """
    east1, north1 = pixel_to_utm(img_bbox[0], img_bbox[1], transform)
    east2, north2 = pixel_to_utm(img_bbox[2], img_bbox[3], transform)
    return [east1, north1, east2, north2]


def wgs84_to_utm(lat, lng, utm_epsg):
    """ Convert WGS84 coordinates to UTM

    :param lat: latitude in WGS84 system
    :type lat: float
    :param lng: longitude in WGS84 system
    :type lng: float
    :param utm_epsg: UTM coordinate reference system enum constants
    :type utm_epsg: constants.CRS
    :return: east, north coordinates in UTM system
    :rtype: float, float
    """
    utm_proj = pyproj.Proj(init=CRS.ogc_string(utm_epsg))
    east, north = pyproj.transform(WGS84_PROJ, utm_proj, lng, lat)
    return east, north


def to_wgs84(east, north, epsg):
    """ Convert any CRS with (east, north) coordinates to WGS84

    :param east: east coordinate
    :type east: float
    :param north: north coordinate
    :type north: float
    :param epsg: CRS enum constants
    :type epsg: constants.CRS
    :return: latitude and longitude coordinates in WGS84 system
    :rtype: float, float
    """
    projection = pyproj.Proj(init=CRS.ogc_string(epsg))
    lng, lat = pyproj.transform(projection, WGS84_PROJ, east, north)
    return lat, lng


def utm_to_wgs84(east, north, utm_epsg):
    """ Convert UTM coordinates to WGS84

    :param east: east coordinate in UTM system
    :type east: float
    :param north: north coordinate in UTM system
    :type north: float
    :param utm_epsg: UTM coordinate reference system enum constants
    :type utm_epsg: constants.CRS
    :return: latitude and longitude coordinates in WGS84 system
    :rtype: float, float
    """
    return to_wgs84(east, north, utm_epsg)


def utm_to_pixel(east, north, transform, truncate=True):
    """ Convert UTM coordinate to image coordinate given a transform

    :param east: east coordinate of point
    :type east: float
    :param north: north coordinate of point
    :type north: float
    :param transform: georeferencing transform
    :type transform: list
    :param truncate: Whether to truncate pixel coordinates. Default is `True`
    :type truncate: bool
    :return: row and column pixel image coordinates
    :rtype: float, float or int, int
    """
    column = (east - transform[0]) / transform[1]
    row = (north - transform[3]) / transform[5]
    if truncate:
        return int(row + ERR), int(column + ERR)
    return row, column


def pixel_to_utm(row, column, transform):
    """ Convert pixel coordinate to UTM coordinate given a transform

    :param row: row pixel coordinate
    :type row: int or float
    :param column: column pixel coordinate
    :type column: int or float
    :param transform: georeferencing transform
    :type transform: list
    :return: east, north UTM coordinates
    :rtype: float, float
    """
    east = transform[0] + column * transform[1]
    north = transform[3] + row * transform[5]
    return east, north


def wgs84_to_pixel(lat, lng, utm_epsg, transform, truncate=True):
    """ Convert WGS84 coordinates to pixel image coordinates given a UTM CRS and a transform

    :param lat: latitude of point
    :type lat: float
    :param lng: longitude of point
    :type lng: float
    :param utm_epsg: UTM coordinate reference system enum constants
    :type utm_epsg: constants.CRS
    :param transform: georeferencing transform
    :type transform: list
    :param truncate: Whether to truncate pixel coordinates. Default is `True`
    :type truncate: bool
    :return: row and column pixel image coordinates
    :rtype: float, float or int, int
    """
    east, north = wgs84_to_utm(lat, lng, utm_epsg)
    row, column = utm_to_pixel(east, north, transform, truncate=truncate)
    return row, column


def get_utm_epsg_from_latlon(lat, lon):
    """ Get CRS for UTM zone in which (lat,lon) is contained.

    :param lat: latitude
    :type lat: float
    :param lon: longitude
    :type lon: float
    :return: CRS of the zone containing the lat,lon point
    :rtype: constants.CRS
    """
    return CRS.get_utm_from_wgs84(lat, lon)


def transform_point(point, src_crs, tht_crs):
    """ Maps point form src_crs to tgt_crs

    :param point: a tuple (x,y)
    :type point: tuple[float] of length 2
    :param src_crs: source CRS
    :type src_crs: constants.CRS
    :param tht_crs: target CRS
    :type tht_crs: constants.CRS
    :return: point in target CRS
    :rtype: tuple[float]
    """
    raise NotImplementedError


def transform_bbox(bbox, tgt_crs):
    """ Maps bbox from bbox.get_crs() to tgt_crs

    :param bbox: bounding box
    :type bbox: common.BBox
    :param tgt_crs: target CRS
    :type tgt_crs: constants.CRS
    :return: bounding box in target CRS
    :rtype: common.BBox
    """
    raise NotImplementedError
